import os
import copy
from functools import partial
import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor

import numpy as np

from utils import xml_util

from algorithm.frap_pub.agent import Agent
from algorithm.frap_pub.definitions import ROOT_DIR


class PlanningOnlyAgent(Agent):

    def __init__(self):

        self.rng = np.random.Generator(np.random.MT19937(23423))

        self.env = None

        self.phases = ['0S_2S_0L_2L', '1S_3S_1L_3L']
        self.planning_iterations = 10
        self.action_sampling_size = 2

        context = mp.get_context('spawn')
        self.process_pool_executor = ProcessPoolExecutor(max_workers=4, mp_context=context)

        self.previous_actions = [None] * len(self.phases)

        xml_util.register_copyreg()

    def set_simulation_environment(self, env):
        self.env = env

    def choose_action(self, step, state, *args, **kwargs):

        intersection_index = kwargs.get('intersection_index', None)

        if intersection_index is None:
            raise ValueError('intersection_index must be declared')

        action = self._choose_action(step, state, intersection_index)

        self.previous_actions[intersection_index] = action

        return action

    def _expand_data_list(self, data, deepcopy=False):
        for i in range(len(data) - 1, -1, -1):
            if deepcopy:
                data = data[:i] + [copy.deepcopy(data[i]) for _ in range(self.action_sampling_size)] + data[i + 1:]
            else:
                data = data[:i] + [data[i]] * self.action_sampling_size + data[i + 1:]

        return data

    def _get_possible_actions(self, state, intersection_index):

        possible_actions = self.rng.choice(range(2), self.action_sampling_size, replace=False)

        return possible_actions

    def _choose_action(self, step, state, intersection_index):

        previous_planning_actions = []
        if self.previous_actions[intersection_index] is not None:
            previous_planning_actions.append(self.previous_actions[intersection_index])

        save_state_filepath = self.env.save_state()

        states = [state]
        planning_step_list = [step]
        previous_planning_actions_list = [[]]
        save_state_filepath_list = [save_state_filepath]
        possible_future_rewards = np.array([])

        remaining_planning_iterations = self.planning_iterations

        simulation_possibility_kwargs = {
            'env': self.env,
            'original_step': step,
            'intersection_index': intersection_index
        }

        while remaining_planning_iterations > 0:

            save_state_filepath_to_remove_list = save_state_filepath_list

            possible_actions = []

            for state in states:
                possible_actions.extend(self._get_possible_actions(state, intersection_index))

            save_state_filepath_list = self._expand_data_list(save_state_filepath_list)

            planning_step_list = self._expand_data_list(planning_step_list)

            previous_planning_actions_list = self._expand_data_list(previous_planning_actions_list, deepcopy=True)

            states, rewards, steps_iterated_list, save_state_filepath_list = list(zip(*list(
                self.process_pool_executor.map(
                    partial(PlanningOnlyAgent._run_simulation_possibility, **simulation_possibility_kwargs),
                    possible_actions, planning_step_list, previous_planning_actions_list, save_state_filepath_list
                ))))

            for save_state_filepath_to_remove in save_state_filepath_to_remove_list:
                os.remove(save_state_filepath_to_remove)

            if len(possible_future_rewards) != 0:
                for i in range(len(possible_future_rewards) - 1, -1, -1):
                    possible_future_rewards = \
                        np.concatenate([possible_future_rewards[:i],
                                        [possible_future_rewards[i]] * self.action_sampling_size,
                                        possible_future_rewards[i + 1:]],
                                       axis=None)
            else:
                possible_future_rewards = np.array([0] * self.action_sampling_size)

            possible_future_rewards = np.add(possible_future_rewards, rewards)

            if len(previous_planning_actions_list) != 0:
                for i in range(len(previous_planning_actions_list) - 1, -1, -1):
                    previous_planning_actions_list[i].append(possible_actions[i])

            save_state_filepath_list = list(save_state_filepath_list)

            for i in range(len(planning_step_list) - 1, -1, -1):
                planning_step_list[i] += steps_iterated_list[i]

            remaining_planning_iterations -= 1

        best_actions = np.flatnonzero(possible_future_rewards == np.max(possible_future_rewards))

        if len(best_actions) > 1:
            action = self.rng.choice(best_actions)
        else:
            action = best_actions[0]

        return action

    @staticmethod
    def _run_simulation_possibility(
            action,
            planning_step,
            previous_planning_actions,
            save_state_filepath,
            env,
            original_step,
            intersection_index):

        try:

            env = copy.deepcopy(env)

            env.external_configurations['SUMOCFG_PARAMETERS'].update(
                {
                    '--begin': planning_step,
                    '--load-state': save_state_filepath
                }
            )

            execution_name = 'planning_for_step' + '_' + str(original_step) + '__' + \
                             'previous_phases' + '_' + (
                                 '-'.join(str(x) for x in previous_planning_actions)
                                 if len(previous_planning_actions) else str(None)) + '__' + \
                             'planning_step' + '_' + str(planning_step) + '__' + \
                             'phase' + '_' + str(action)

            env.path_to_log = env.path_to_log + '__' + execution_name
            if not os.path.exists(ROOT_DIR + '/' + env.path_to_log):
                os.makedirs(ROOT_DIR + '/' + env.path_to_log)

            _, next_action = env.reset_for_planning(execution_name)

            action_list = ['no_op'] * len(next_action)
            action_list[intersection_index] = action

            next_state, reward, done, steps_iterated, _ = env.step(action_list)
            next_state = next_state[0]
            reward = reward[0]

            save_state_filepath = env.save_state()

            env.end_sumo()

        except Exception as e:
            print(e)
            raise e

        return next_state, reward, steps_iterated, save_state_filepath
